@@ -1,4 +1,5 @@
 from __future__ import division
+import collections
 from collections import OrderedDict
 try:
     from collections.abc import Iterable, Mapping, MutableMapping
@@ -10,14 +11,13 @@ import functools
 import itertools
 import json
 import math
-import threading
 import time
 import warnings
 
 try:
-    from bson import json_util, SON, BSON
+    from bson import json_util, SON, BSON, Timestamp
 except ImportError:
-    json_utils = SON = BSON = None
+    json_utils = SON = BSON = Timestamp = None
 try:
     import execjs
 except ImportError:
@@ -36,12 +36,11 @@ except ImportError:
         BEFORE = False
         AFTER = True
 
-    _READ_PREFERENCE_PRIMARY = None
+    from mongomock.read_preferences import PRIMARY as _READ_PREFERENCE_PRIMARY
 
 from sentinels import NOTHING
 from six import iteritems
 from six import iterkeys
-from six import itervalues
 from six import string_types
 from six import text_type
 
@@ -49,6 +48,7 @@ from six import text_type
 from mongomock import aggregate
 from mongomock import ConfigurationError, DuplicateKeyError, BulkWriteError
 from mongomock.filtering import filter_applies
+from mongomock.filtering import iter_key_candidates
 from mongomock.filtering import resolve_key
 from mongomock.filtering import resolve_sort_key
 from mongomock import helpers
@@ -63,7 +63,22 @@ from mongomock.results import UpdateResult
 from mongomock.write_concern import WriteConcern
 from mongomock import WriteError
 
-lock = threading.RLock()
+_KwargOption = collections.namedtuple('KwargOption', ['typename', 'default', 'attrs'])
+
+_WITH_OPTIONS_KWARGS = {
+    'codec_options': _KwargOption(
+        'bson.codec_options.CodecOptions', None,
+        ('document_class', 'tz_aware', 'uuid_representation')),
+    'read_preference': _KwargOption(
+        'pymongo.read_preference.ReadPreference', _READ_PREFERENCE_PRIMARY,
+        ('document', 'mode', 'mongos_mode', 'max_staleness')),
+    'write_concern': _KwargOption(
+        'pymongo.write_concern.WriteConcern', WriteConcern(),
+        ('acknowledged', 'document', 'is_server_default')),
+    'read_concern': _KwargOption(
+        'pymongo.read_concern.ReadConcern', None,
+        ('document', 'level', 'ok_for_legacy'))
+}
 
 
 def validate_is_mapping(option, value):
@@ -149,10 +164,13 @@ class BulkWriteOperation(object):
             if result.get('upserted'):
                 ret_val['upserted'] = result.get('upserted')
                 ret_val['nUpserted'] = result.get('n')
+            else:
+                matched = result.get('n')
+                if matched is not None:
+                    ret_val['nMatched'] = matched
             modified = result.get('nModified')
             if modified is not None:
                 ret_val['nModified'] = modified
-                ret_val['nMatched'] = modified
             if result.get('err'):
                 ret_val['err'] = result.get('err')
             return ret_val
@@ -282,9 +300,19 @@ class BulkOperationBuilder(object):
         has_update = False
         has_insert = False
         broken_nModified_info = False
-        for execute_func in self.executors:
+        for index, execute_func in enumerate(self.executors):
             exec_name = execute_func.__name__
-            op_result = execute_func()
+            try:
+                op_result = execute_func()
+            except WriteError as error:
+                result['writeErrors'].append({
+                    'index': index,
+                    'code': error.code,
+                    'errmsg': str(error),
+                })
+                if self.ordered:
+                    break
+                continue
             for (key, value) in op_result.items():
                 self.__aggregate_operation_result(result, key, value)
             if exec_name == 'exec_update':
@@ -303,12 +331,16 @@ class BulkOperationBuilder(object):
             pass
         else:
             result.pop('nModified')
+
+        if result.get('writeErrors'):
+            raise BulkWriteError(result)
+
         return result
 
     def add_insert(self, doc):
         self.insert(doc)
 
-    def add_update(self, selector, doc, multi, upsert, collation=None,
+    def add_update(self, selector, doc, multi=False, upsert=False, collation=None,
                    array_filters=None):
         if array_filters:
             raise NotImplementedError(
@@ -327,18 +359,12 @@ class BulkOperationBuilder(object):
 
 class Collection(object):
 
-    def __init__(self, database, name, create=False, write_concern=None):
-        self.name = name
-        self.full_name = '{0}.{1}'.format(database.name, name)
+    def __init__(self, database, name, _db_store, write_concern=None, read_preference=None):
         self.database = database
-        self._documents = OrderedDict()
-        self._force_created = create
+        self._name = name
+        self._db_store = _db_store
         self._write_concern = write_concern or WriteConcern()
-        self._uniques = {}
-        self._index_information = {'_id_': {'v': 2, 'key': [('_id', 1)], 'ns': self.full_name}}
-
-    def _is_created(self):
-        return self._documents or self._uniques or self._force_created
+        self._read_preference = read_preference or _READ_PREFERENCE_PRIMARY
 
     def __repr__(self):
         return "Collection({0}, '{1}')".format(self.database, self.name)
@@ -353,10 +379,22 @@ class Collection(object):
                 (self.__class__.__name__, attr, self.name, attr, self.name, attr))
         return self.__getitem__(attr)
 
+    @property
+    def full_name(self):
+        return '{0}.{1}'.format(self.database.name, self._name)
+
+    @property
+    def name(self):
+        return self._name
+
     @property
     def write_concern(self):
         return self._write_concern
 
+    @property
+    def read_preference(self):
+        return self._read_preference
+
     def initialize_unordered_bulk_op(self):
         return BulkOperationBuilder(self, ordered=False)
 
@@ -382,6 +420,10 @@ class Collection(object):
             validate_is_mutable_mapping('document', document)
         return InsertManyResult(self._insert(documents, session), acknowledged=True)
 
+    @property
+    def _store(self):
+        return self._db_store[self._name]
+
     def _insert(self, data, session=None):
         if session:
             raise NotImplementedError('Mongomock does not handle sessions yet')
@@ -390,12 +432,12 @@ class Collection(object):
             for index, item in enumerate(data):
                 try:
                     results.append(self._insert(item))
-                except DuplicateKeyError:
+                except WriteError as error:
                     raise BulkWriteError({
                         'writeErrors': [{
                             'index': index,
-                            'code': 11000,
-                            'errmsg': 'E11000 duplicate key error',
+                            'code': error.code,
+                            'errmsg': str(error),
                             'op': item,
                         }],
                         'nInserted': index,
@@ -419,21 +461,24 @@ class Collection(object):
         object_id = data['_id']
         if isinstance(object_id, dict):
             object_id = helpers.hashdict(object_id)
-        if object_id in self._documents:
+        if object_id in self._store:
             raise DuplicateKeyError('E11000 Duplicate Key Error', 11000)
-        with lock:
-            self._documents[object_id] = self._internalize_dict(data)
+        self._store[object_id] = self._internalize_dict(data)
         try:
             self._ensure_uniques(data)
         except DuplicateKeyError:
             # Rollback
-            del self._documents[object_id]
+            del self._store[object_id]
             raise
         return data['_id']
 
     def _ensure_uniques(self, new_data):
         # Note we consider new_data is already inserted in db
-        for unique, is_sparse in self._uniques.values():
+        for index in self._store.indexes.values():
+            if not index.get('unique'):
+                continue
+            unique = index.get('key')
+            is_sparse = index.get('sparse')
             find_kwargs = {}
             for key, _ in unique:
                 try:
@@ -543,10 +588,6 @@ class Collection(object):
                         existing_document, v, spec, _set_updater, subdocument)
 
                 elif k == '$currentDate':
-                    for value in itervalues(v):
-                        if value == {'$type': 'timestamp'}:
-                            raise NotImplementedError('timestamp is not supported so far')
-
                     subdocument = self._update_document_fields_with_positional_awareness(
                         existing_document, v, spec, _current_date_updater, subdocument)
 
@@ -667,7 +708,25 @@ class Collection(object):
                             subdocument[field] = []
                         push_results = subdocument[field]
                         if isinstance(value, dict) and '$each' in value:
-                            push_results += list(value['$each'])
+                            if '$position' in value:
+                                push_results = \
+                                    push_results[0:value['$position']] + \
+                                    list(value['$each']) + \
+                                    push_results[value['$position']:]
+                            else:
+                                push_results += list(value['$each'])
+
+                            if '$sort' in value:
+                                sort_spec = value['$sort']
+                                if isinstance(sort_spec, dict):
+                                    sort_key = set(sort_spec.keys()).pop()
+                                    push_results = sorted(
+                                        push_results,
+                                        key=lambda d: helpers.get_value_by_dot(d, sort_key),
+                                        reverse=set(sort_spec.values()).pop() < 0)
+                                else:
+                                    push_results = sorted(push_results, reverse=sort_spec < 0)
+
                             if '$slice' in value:
                                 slice_value = value['$slice']
                                 if slice_value < 0:
@@ -676,6 +735,12 @@ class Collection(object):
                                     push_results = []
                                 else:
                                     push_results = push_results[:slice_value]
+
+                            unused_modifiers = \
+                                set(value.keys()) - {'$each', '$slice', '$position', '$sort'}
+                            if unused_modifiers:
+                                raise WriteError(
+                                    'Unrecognized clause in $push: ' + unused_modifiers.pop())
                         else:
                             push_results.append(value)
                         subdocument[field] = push_results
@@ -689,7 +754,7 @@ class Collection(object):
                                 raise ValueError('field names cannot start with $ [{}]'.format(k))
                         _id = spec.get('_id', existing_document.get('_id'))
                         existing_document.clear()
-                        if _id:
+                        if _id is not None:
                             existing_document['_id'] = _id
                         if BSON:
                             # bson validation
@@ -716,14 +781,24 @@ class Collection(object):
                 upserted_id = self._insert(existing_document)
                 num_updated += 1
             elif existing_document != original_document_snapshot:
-                # Document has been modified in-place, last thing is to make sure it
-                # still respect the unique indexes and if not to revert modifications
+                # Document has been modified in-place.
+
+                # Make sure the ID was not change.
+                if original_document_snapshot.get('_id') != existing_document.get('_id'):
+                    # Rollback.
+                    self._store[original_document_snapshot['_id']] = original_document_snapshot
+                    raise WriteError(
+                        "After applying the update, the (immutable) field '_id' was found to have "
+                        'been altered to _id: {}'.format(existing_document.get('_id')))
+
+                # Make sure it still respect the unique indexes and, if not, to
+                # revert modifications
                 try:
                     self._ensure_uniques(existing_document)
                     num_updated += 1
                 except DuplicateKeyError:
-                    # Rollback
-                    self._documents[original_document_snapshot['_id']] = original_document_snapshot
+                    # Rollback.
+                    self._store[original_document_snapshot['_id']] = original_document_snapshot
                     raise
 
             if not multi:
@@ -837,22 +912,28 @@ class Collection(object):
     def find(self, filter=None, projection=None, skip=0, limit=0,
              no_cursor_timeout=False, cursor_type=None, sort=None,
              allow_partial_results=False, oplog_replay=False, modifiers=None,
-             batch_size=0, manipulate=True, collation=None, session=None):
+             batch_size=0, manipulate=True, collation=None, session=None,
+             max_time_ms=None):
         spec = filter
         if spec is None:
             spec = {}
         validate_is_mapping('filter', spec)
-        return Cursor(self, spec, sort, projection, skip, limit, collation=collation)
+        return Cursor(self, spec, sort, projection, skip, limit,
+                      collation=collation).max_time_ms(max_time_ms)
 
     def _get_dataset(self, spec, sort, fields, as_class):
-        dataset = (self._copy_only_fields(document, fields, as_class)
-                   for document in self._iter_documents(spec))
+        dataset = self._iter_documents(spec)
         if sort:
-            for sortKey, sortDirection in reversed(sort):
+            for sort_key, sort_direction in reversed(sort):
+                if sort_key == '$natural':
+                    if sort_direction < 0:
+                        dataset = iter(reversed(list(dataset)))
+                    continue
                 dataset = iter(sorted(
-                    dataset, key=lambda x: resolve_sort_key(sortKey, x),
-                    reverse=sortDirection < 0))
-        return dataset
+                    dataset, key=lambda x: resolve_sort_key(sort_key, x),
+                    reverse=sort_direction < 0))
+        for document in dataset:
+            yield self._copy_only_fields(document, fields, as_class)
 
     def _copy_field(self, obj, container):
         if isinstance(obj, list):
@@ -1049,10 +1130,10 @@ class Collection(object):
 
     def _iter_documents(self, filter=None):
         # Validate the filter even if no documents can be returned.
-        if not self._documents:
+        if self._store.is_empty:
             filter_applies(filter, {})
 
-        return (document for document in list(itervalues(self._documents))
+        return (document for document in list(self._store.documents)
                 if filter_applies(filter, document))
 
     def find_one(self, filter=None, *args, **kwargs):  # pylint: disable=keyword-arg-before-vararg
@@ -1168,7 +1249,7 @@ class Collection(object):
             doc_id = doc['_id']
             if isinstance(doc_id, dict):
                 doc_id = helpers.hashdict(doc_id)
-            del self._documents[doc_id]
+            del self._store[doc_id]
             deleted_count += 1
             if not multi:
                 break
@@ -1196,7 +1277,7 @@ class Collection(object):
         if kwargs.pop('session', None):
             raise NotImplementedError('Mongomock does not handle sessions yet')
         if filter is None:
-            return len(self._documents)
+            return len(self._store)
         return len(list(self._iter_documents(filter)))
 
     def count_documents(self, filter, **kwargs):
@@ -1252,17 +1333,15 @@ class Collection(object):
         is_sparse = kwargs.pop('sparse', False)
 
         index_string = helpers.gen_index_name(index_list)
-        self._index_information[index_string] = {
-            'key': index_list,
-            'ns': self.full_name,
-            'v': 2,
-        }
+        index_dict = {'key': index_list}
+        if is_sparse:
+            index_dict['sparse'] = True
 
         # Check that documents already verify the uniquess of this new index.
         if is_unique:
-            self._index_information[index_string]['unique'] = True
+            index_dict['unique'] = True
             indexed = set()
-            for doc in itervalues(self._documents):
+            for doc in self._store.documents:
                 index = []
                 for key, unused_order in index_list:
                     try:
@@ -1271,12 +1350,14 @@ class Collection(object):
                         if is_sparse:
                             continue
                         index.append(None)
+                if is_sparse and not index:
+                    continue
                 index = tuple(index)
                 if index in indexed:
                     raise DuplicateKeyError('E11000 Duplicate Key Error', 11000)
                 indexed.add(index)
 
-            self._uniques[index_string] = (index_list, is_sparse)
+        self._store.indexes[index_string] = index_dict
 
         return index_string
 
@@ -1302,31 +1383,44 @@ class Collection(object):
         else:
             name = index_or_name
         try:
-            del self._index_information[name]
+            del self._store.indexes[name]
         except KeyError:
             raise OperationFailure('index not found with name [%s]' % name)
-        self._uniques.pop(name, None)
 
     def drop_indexes(self, session=None):
         if session:
             raise NotImplementedError('Mongomock does not handle sessions yet')
-        self._uniques = {}
-        self._index_information = {'_id_': {'v': 2, 'key': [('_id', 1)], 'ns': self.full_name}}
+        self._store.indexes = {}
 
     def reindex(self, session=None):
         if session:
             raise NotImplementedError('Mongomock does not handle sessions yet')
 
+    def _list_all_indexes(self):
+        if not self._store.is_created:
+            return
+        yield '_id_', {'key': [('_id', 1)]}
+        for name, information in self._store.indexes.items():
+            yield name, information
+
     def list_indexes(self, session=None):
         if session:
             raise NotImplementedError('Mongomock does not handle sessions yet')
-        for name, information in self._index_information.items():
-            yield dict(information, key=dict(information['key']), name=name)
+        for name, information in self._list_all_indexes():
+            yield dict(
+                information,
+                key=dict(information['key']),
+                name=name,
+                v=2,
+                ns=self.full_name)
 
     def index_information(self, session=None):
         if session:
             raise NotImplementedError('Mongomock does not handle sessions yet')
-        return copy.deepcopy(self._index_information)
+        return {
+            name: dict(index, v=2, ns=self.full_name)
+            for name, index in self._list_all_indexes()
+        }
 
     def map_reduce(self, map_func, reduce_func, out, full_response=False,
                    query=None, limit=0, session=None):
@@ -1501,34 +1595,36 @@ class Collection(object):
         return aggregate.process_pipeline(in_collection, self.database, pipeline, session)
 
     def with_options(self, **kwargs):
-        default_kwargs = {
-            'codec_options': None,
-            'read_preference': _READ_PREFERENCE_PRIMARY,
-            'write_concern': WriteConcern(),
-            'read_concern': None,
-        }
-        forbidden_kwargs = set(kwargs.keys()) - set(default_kwargs)
+        forbidden_kwargs = set(kwargs.keys()) - set(_WITH_OPTIONS_KWARGS)
         if forbidden_kwargs:
             raise TypeError(
                 "with_options() got an unexpected keyword argument '%s'" % forbidden_kwargs.pop())
-        for key, default_value in iteritems(default_kwargs):
-            value = kwargs.get(key)
-            if value is not None and default_value != value:
+
+        for key, value in iteritems(kwargs):
+            if value is None:
+                continue
+            options = _WITH_OPTIONS_KWARGS[key]
+            for attr in options.attrs:
+                if not hasattr(value, attr):
+                    raise TypeError(
+                        '{} must be an instance of {}'.format(key, options.typename))
+            if key not in ('read_preference', 'write_concern') and options.default != value:
                 raise NotImplementedError(
                     '%s is a valid parameter for with_options but it is currently not implemented '
                     'in Mongomock' % key)
-        return self
+
+        write_concern = kwargs.get('write_concern', self.write_concern)
+        read_preference = kwargs.get('read_preference', self.read_preference)
+        return Collection(
+            self.database, self.name, write_concern=write_concern, read_preference=read_preference,
+            _db_store=self._db_store)
 
     def rename(self, new_name, session=None, **kwargs):
         if session:
             raise NotImplementedError('Mongomock does not handle sessions yet')
-        self.database.rename_collection(self.name, new_name, **kwargs)
+        return self.database.rename_collection(self.name, new_name, **kwargs)
 
     def bulk_write(self, requests, ordered=True, bypass_document_validation=False, session=None):
-        if not ordered:
-            raise NotImplementedError(
-                'Unordered mode is a valid MongoDB operation; however Mongomock'
-                ' does not support it yet.')
         if bypass_document_validation:
             raise NotImplementedError(
                 'Skipping document validation is a valid MongoDB operation;'
@@ -1537,7 +1633,7 @@ class Collection(object):
             raise NotImplementedError(
                 'Sessions are valid in MongoDB 3.6 and newer; however Mongomock'
                 ' does not support them yet.')
-        bulk = BulkOperationBuilder(self)
+        bulk = BulkOperationBuilder(self, ordered=ordered)
         for operation in requests:
             operation._add_to_bulk(bulk)
         return BulkWriteResult(bulk.execute(), True)
@@ -1556,8 +1652,8 @@ class Cursor(object):
         self._skip = skip
         self._factory_last_generated_results = None
         self._results = None
-        self._factory = functools.partial(collection._get_dataset,
-                                          spec, sort, projection, dict)
+        self._factory = functools.partial(
+            collection._get_dataset, spec, sort, projection, dict)
         # pymongo limit defaults to 0, returning everything
         self._limit = limit if limit != 0 else None
         self._collation = collation
@@ -1605,27 +1701,9 @@ class Cursor(object):
         self._emitted = 0
 
     def sort(self, key_or_list, direction=None):
-        if direction is None:
-            direction = 1
-
-        def _make_sort_factory_layer(upper_factory, sort_key, sort_direction):
-            if sort_key == '$natural':
-                if sort_direction >= 0:
-                    return upper_factory
-                return lambda: reversed(list(upper_factory()))
-
-            def layer():
-                return sorted(upper_factory(), key=lambda x: resolve_sort_key(sort_key, x),
-                              reverse=sort_direction < 0)
-            return layer
-
-        if isinstance(key_or_list, dict):
-            key_or_list = list(key_or_list.items())
-        if isinstance(key_or_list, (tuple, list)):
-            for sort_key, sort_direction in reversed(key_or_list):
-                self._factory = _make_sort_factory_layer(self._factory, sort_key, sort_direction)
-        else:
-            self._factory = _make_sort_factory_layer(self._factory, key_or_list, direction)
+        self._sort = helpers.create_index_list(key_or_list, direction)
+        self._factory = functools.partial(
+            self.collection._get_dataset, self._spec, self._sort, self._projection, dict)
         return self
 
     def count(self, with_limit_and_skip=False):
@@ -1646,6 +1724,14 @@ class Cursor(object):
     def close(self):
         pass
 
+    def hint(self, unused_hint):
+        if self._emitted:
+            raise InvalidOperation('cannot set options after executing query')
+        # TODO(pascal): Once we implement $text indexes and queries, raise an
+        # exception if hint is used on a $text query.
+        # https://docs.mongodb.com/manual/reference/method/cursor.hint/#behavior
+        return self
+
     def distinct(self, key, session=None):
         if session:
             raise NotImplementedError('Mongomock does not handle sessions yet')
@@ -1654,17 +1740,17 @@ class Cursor(object):
         unique = set()
         unique_dict_vals = []
         for x in self._compute_results():
-            value = resolve_key(key, x)
-            if value == NOTHING:
-                continue
-            if isinstance(value, dict):
-                if any(dict_val == value for dict_val in unique_dict_vals):
+            for value in iter_key_candidates(key, x):
+                if value == NOTHING:
                     continue
-                unique_dict_vals.append(value)
-            else:
-                unique.update(
-                    value if isinstance(
-                        value, (tuple, list)) else [value])
+                if isinstance(value, dict):
+                    if any(dict_val == value for dict_val in unique_dict_vals):
+                        continue
+                    unique_dict_vals.append(value)
+                else:
+                    unique.update(
+                        value if isinstance(
+                            value, (tuple, list)) else [value])
         return list(unique) + unique_dict_vals
 
     def __getitem__(self, index):
@@ -1692,12 +1778,11 @@ class Cursor(object):
             self._skip = skip
             self._limit = limit
             return self
-        elif not isinstance(index, int):
+        if not isinstance(index, int):
             raise TypeError("index '%s' cannot be applied to Cursor instances" % index)
-        elif index < 0:
+        if index < 0:
             raise IndexError('Cursor instances do not support negativeindices')
-        else:
-            return self._compute_results(with_limit_and_skip=True)[index]
+        return self._compute_results(with_limit_and_skip=True)[index]
 
     def __enter__(self):
         return self
@@ -1709,6 +1794,12 @@ class Cursor(object):
     def alive(self):
         return self._emitted != self.count()
 
+    def max_time_ms(self, max_time_ms):
+        if max_time_ms is not None and not isinstance(max_time_ms, int):
+            raise TypeError('max_time_ms must be an integer or None')
+        # Currently the value is ignored as mongomock never times out.
+        return self
+
 
 def _set_updater(doc, field_name, value):
     if isinstance(value, (tuple, list)):
@@ -1759,9 +1850,27 @@ def _min_updater(doc, field_name, value):
         doc[field_name] = min(doc.get(field_name, value), value)
 
 
+_LAST_TIMESTAMP_INC = []
+
+
+def _get_current_timestamp():
+    if not Timestamp:
+        raise NotImplementedError('timestamp is not supported. Import pymongo to use it.')
+    now = int(time.time())
+    if _LAST_TIMESTAMP_INC and _LAST_TIMESTAMP_INC[0] == now:
+        _LAST_TIMESTAMP_INC[1] += 1
+    else:
+        del _LAST_TIMESTAMP_INC[:]
+        _LAST_TIMESTAMP_INC.extend([now, 1])
+    return Timestamp(now, _LAST_TIMESTAMP_INC[1])
+
+
 def _current_date_updater(doc, field_name, value):
     if isinstance(doc, dict):
-        doc[field_name] = datetime.utcnow()
+        if value == {'$type': 'timestamp'}:
+            doc[field_name] = _get_current_timestamp()
+        else:
+            doc[field_name] = datetime.utcnow()
 
 
 _updaters = {
